#!/usr/local/bin/env python
from __future__ import division
import sys
from future.utils import listvalues
from copy import deepcopy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from radd.tools import colors, messages, analyze
from radd.tools.analyze import bootstrap_data
from radd import theta
from IPython.display import display, Latex
from ipywidgets import IntProgress, HTML, Box


def pandaify_results(gort, ssrt, tb=.68, bootstrap=True, bootinfo={'nsubjects':25, 'ntrials':1000, 'groups':['ssd']}, ssd=np.array([[.2, .25, .3, .35, .4]]), clmap=None):
    nl, nssd, nssPer = ssrt.shape
    nl, ntrials = gort.shape
    nss = nssd * nssPer
    dfColumns=['cond', 'ttype', 'ssd', 'response', 'acc', 'rt', 'ssrt', 'trial']
    if clmap is not None:
        levelNames = np.sort(np.hstack(listvalues(clmap)).tolist())
    else:
        levelNames = ['flat']*nl
    # bootinfo['groups'] = bootinfo['groups'] + list(clmap)
    dfIndex = np.arange(ntrials)
    dfList = []
    for i in range(nl):
        ert = gort[i, :nss].reshape(ssrt[i].shape)
        goTrialOutcomes = np.hstack(np.where(gort[i, nss:] < tb, 1, 0))
        # ssTrialOutcomes = np.hstack(np.hstack(np.where(ert <= ssrt[i], 1, 0)))
        ssTrialOutcomes = np.hstack(np.hstack(np.where(ssrt[i] <= ert, 0, 1)))
        delays = np.append(np.hstack([[delay]*nssPer for delay in ssd[i]]), [1000]*nss)
        response = np.append(ssTrialOutcomes, goTrialOutcomes)
        responseTime = gort[i]
        ssResponseTime = np.append(np.hstack(np.hstack(ssrt[i])), [np.nan]*nss)
        ttype = np.append(np.zeros(nss), np.ones(nss))
        acc = np.where(ttype == response, 1, 0)
        cond = [levelNames[i]]*ntrials
        trial = np.arange(1, ntrials+1)
        dfData = [cond, ttype, delays, response, acc, responseTime, ssResponseTime, trial]
        df = pd.DataFrame(dict(zip(dfColumns, dfData)), index=dfIndex)
        dfList.append(df[dfColumns])
    resultsdf = pd.concat(dfList)
    resultsdf.reset_index(drop=True, inplace=True)
    if bootstrap:
        resultsdf = bootstrap_data(resultsdf, nsubjects=bootinfo['nsubjects'], n=bootinfo['ntrials'], groups=bootinfo['groups'])
    return resultsdf



class PBinJ(object):
    """ initialize multiple progress bars for tracking nested stages of fitting routine
    """
    def __init__(self, n=1, value=0, status='{}', color='b', width='60%', height='22px'):
        self.displayed = False
        self.style_bar(n=n, value=value, status=status, color=color, width=width, height=height)

    def style_bar(self, n=1, value=0, status='{}', color='b', width='60%', height='22px'):
        colordict = {'g': 'success', 'b': '', 'r': 'danger', 'y': 'warning', 'c': 'info'}
        bar_style = colordict[color]
        self.bar = IntProgress(min=0, max=n, value=value, bar_style=bar_style)
        self.status = status
        self.bar.bar_style = bar_style
        self.bar.width = width
        self.bar.height = height

    def reset_bar(self, color=False):
        self.update(value=0)

    def update(self, value=None, status=None):
        if not self.displayed:
            display(self.bar)
            self.displayed=True
        if status is not None:
            if hasattr(status, '__iter__'):
                status = self.status.format(*status)
            else:
                status = self.status.format(status)
            self.bar.description = status
        if value is not None:
            self.bar.value = value+1

    def clear(self):
        self.bar.close()
        self.displayed = False


class BasinCallback(object):
    """ A callback function for reporting basinhopping status
    Arguments:
        x (array):
            parameter values
        fmin (float):
            function value of the trial minimum, and
        accept (bool):
            whether or not that minimum was accepted
    """
    def __init__(self,  n=1, value=0, status='{:.5fz} / {:.5fz}', color='r', fmin=1000.):
        self.pbar = PBinJ(n=n, value=value, status=status, color=color)
        self.reset(history=True, gbasin=True, fmin=fmin)
        self.xhistory = []

    def reset(self, history=True, bar=False, gbasin=False, get_call=False, fmin=1000.):
        fmin = MyFloat(fmin)
        if history:
            self.history = [fmin]
        if gbasin:
            self.gbasin = fmin
        if bar:
            self.pbar.reset_bar()
        if get_call:
            return self.callback

    def callback(self, x, fmin, accept):
        if fmin <= np.min(self.history) and fmin<=self.gbasin:
            self.gbasin = fmin
            self.xhistory.append((fmin, x))
            self.reset(history=True, bar=True)
        if accept:
            self.history.append(fmin)
            status=(MyFloat(x) for x in [self.gbasin, fmin])
            self.pbar.update(value=len(self.history), status=status)

    def clear(self):
        self.pbar.clear()


class GradientCallback(object):
    """ A callback function for reporting basinhopping status
    Arguments:
        x (array):
            parameter values
        fmin (float):
            function value of the trial minimum, and
        accept (bool):
            whether or not that minimum was accepted
    """
    def __init__(self,  n=1, value=0, status='{:.5fz} / {:.5fz}', color='g', fmin=1000.):
        self.pbar = PBinJ(n=n, value=value, status=status, color=color)
        self.reset(history=True, lbasin=True, fmin=fmin)
        self.xhistory = []
        self.nsame = 0

    def reset(self, history=True, bar=False, lbasin=False, get_call=False, fmin=1000.):
        fmin = MyFloat(fmin)
        if history:
            self.history = [fmin]
            self.nsame = 0
        if lbasin:
            self.lbasin = fmin
        if bar:
            self.pbar.reset_bar()
        if get_call:
            return self.callback

    def callback(self, params, iter, resid):
        fmin = np.sum(resid**2)
        if fmin < np.min(self.history) and fmin<self.lbasin:
            self.history.append(fmin)
            self.lbasin = fmin
            self.xhistory.append((fmin, params.valuesdict()))
            self.nsame = 0
        else:
            self.nsame += 1
            #self.reset(history=True, bar=True)
        status=(MyFloat(x) for x in [self.lbasin, fmin])
        self.pbar.update(value=iter, status=status)

    def clear(self):
        self.pbar.clear()



class MyFloat(float):
    """ remove leading zeros from string formatted floats
    """
    def remove_leading_zero(self, value, string):
        if 1 > value > -1:
            string = string.replace('0', '', 1)
        return string

    def __format__(self, format_string):
        if format_string.endswith('z'):
            format_string = format_string[:-1]
            removezero = True
        else:
            removezero = False
        string = super(MyFloat, self).__format__(format_string)
        return self.remove_leading_zero(self, string) if removezero else string


def rwr(X, get_index=False, n=None):
    """
    Modified from http://nbviewer.ipython.org/gist/aflaxman/6871948
    """
    if isinstance(X, pd.Series):
        X = X.copy()
        X.index = range(len(X.index))
    if n == None:
        n = len(X)
    resample_i = np.floor(np.random.rand(n) * len(X)).astype(int)
    X_resample = (X[resample_i])
    if get_index:
        return resample_i
    else:
        return X_resample

def resample_data(data, n=120, groups=['ssd']):
    """ generates n resampled datasets using rwr()
    for bootstrapping model fits
    """
    df = data.copy()
    bootlist = list()
    if n == None:
        n = len(df)
    for level, level_df in df.groupby(groups):
        boots = level_df.reset_index(drop=True)
        orig_ix = np.asarray(boots.index[:])
        resampled_ix = rwr(orig_ix, get_index=True, n=n)
        bootdf = level_df.irow(resampled_ix)
        bootlist.append(bootdf)
    # concatenate and return all resampled conditions
    return self.model.rangl_data(pd.concat(bootlist))

def extract_popt_fitinfo(finfo=None, plist=None, pcmap=None):
    """ takes optimized dict or DF of vectorized parameters and
    returns dict with only depends_on.keys() containing vectorized vals.
    Is accessed by fit.Optimizer objects after optimization routine.
    ::Arguments::
    finfo (dict/DF):
        finfo is dict if self.fit_on is 'average'
        and DF if self.fit_on is 'subjects' or 'bootstrap'
        contains optimized parameters
    ::Returns::
    popt (dict):
        dict with only depends_on.keys() containing
        vectorized vals
    """
    finfo = dict(deepcopy(finfo))
    plist = list(inits)
    popt = {pkey: finfo[pkey] for pkey in plist}
    for pkey in list(pcmap):
        popt[pkey] = np.array([finfo[pc] for pc in pcmap[pkey]])
    return popt

def params_io(p={}, io='w', iostr='popt'):
    """ read // write parameters dictionaries
    """
    if io == 'w':
        pd.Series(p).to_csv(''.join([iostr, '.csv']))
    elif io == 'r':
        ps = pd.read_csv(''.join([iostr, '.csv']), header=None)
        p = dict(zip(ps[0], ps[1]))
        return p

def fits_io(fitparams, fits=[], io='w', iostr='fits'):
    """ read // write y, wts, yhat arrays
    """
    y = fitparams['y'].flatten()
    wts = fitparams['wts'].flatten()
    fits = fits.flatten()
    if io == 'w':
        index = np.arange(len(fits))
        df = pd.DataFrame({'y': y, 'wts': wts, 'yhat': fits}, index=index)
        df.to_csv(''.join([iostr, '.csv']))
    elif io == 'r':
        df = pd.read_csv(''.join([iostr, '.csv']), index_col=0)
        return df



#!/usr/local/bin/env python
# Authors:
#   Kyle Dunovan (ked64@pitt.edu) &
#   Jeremy Huang (jeremyhuang@cmu.edu)

from __future__ import division
from future.utils import listvalues
import os
import numpy as np
import pandas as pd
from radd import models, theta
from radd.adapt import models_rl
from radd.tools import messages, utils
from copy import deepcopy
from scipy.optimize import basinhopping
from numpy.random import uniform
from lmfit import minimize, fit_report



class BasinBounds(object):
    """ sets conditions for step acceptance during
    basinhopping optimization routine
    Arguments:
        xmin (list): lower boundaries for each parameter
        xmax (list): upper boundaries for each parameter
    """

    def __init__(self, xmin, xmax):
        self.xmin = np.array(xmin)
        self.xmax = np.array(xmax)

    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmin = bool(np.all(x >= self.xmin))
        tmax = bool(np.all(x <= self.xmax))
        return tmin and tmax


class HopStep(object):
    """ scale stepsize of basinhopping optimization according to
    individual parameters ranges (smaller steps for more sensitive params).
    See theta.get_stepsize_scalars() for parameter <--> scalar mapping.
    Arguments:
        keys (list): list of parameter names
        nlevels (list): list of levels per parameter
        stepsize (list): initial stepsize
    """

    def __init__(self, keys, nlevels, stepsize=.05):
        self.stepsize_scalars = self.get_stepsize_scalars(keys, nlevels)
        self.stepsize = stepsize
        self.np = self.stepsize_scalars.size
        self.stepsizeList = []

    def get_stepsize_scalars(self, keys, nlevels):
        """ returns an array of scalars used by for controlling
        stepsize of basinhopping algorithm for each parameter
        """
        scalar_dict={'a': .25,
                    'tr': .2,
                    'v': .35,
                    'ssv': .35,
                    'xb': .2,
                    'si': .01,
                    'z': .1,
                    'sso': .2,
                    'C': .05,
                    'B': .05,
                    'R': .01,
                    'vi': 1.1,
                    'vd': 1.5,
                    'Beta': .1}

        nl = [np.ones(nl) for nl in nlevels]
        stepsize_scalars = np.hstack([scalar_dict[k]*nl[i] for i,k in enumerate(keys)])
        stepsize_scalars = stepsize_scalars.squeeze()
        return stepsize_scalars

    def __call__(self, x):
        s = self.stepsize
        self.stepsizeList.append(s)
        ss = self.stepsize_scalars
        x = np.array([x[i] + uniform(-ss[i]*s, ss[i]*s) for i in range(self.np)])
        return x


def format_local_bounds(xmin, xmax):
    """ groups (xmin, xmax) for each parameter """
    tupler = lambda xlim: tuple([xlim[0], xlim[1]])
    # return map((tupler), zip(xmin, xmax))
    return [tupler(xl) for xl in zip(xmin, xmax)]


def format_basinhopping_bounds(basin_keys, nvary, kind='xdpm'):
    """ returns separate lists of all parameter
    min and max values """
    allbounds = theta.get_bounds(kind=kind)
    xmin, xmax = [], []
    nlevels = [np.ones(nl) for nl in nvary]
    for i, pk in enumerate(basin_keys):
        xmin.append([allbounds[pk][0]] * nlevels[i])
        xmax.append([allbounds[pk][1]] * nlevels[i])
    xmin = np.hstack(xmin).tolist()
    xmax = np.hstack(xmax).tolist()
    return xmin, xmax



class Optimizer(object):
    """ Primary optimization class for handling global and local optimization routines.

    Arguments:
        fitparams (Series): fitparams Series object
        inits (dict): parameter dictionary to initialize model with
        param_sets (list): list of init dicts to run basinhopping over
        basinparams (dict): dictionary of global optimization params
        progress (bool): initialize progress bars (default=True)
        custompath (str): local path from ~ to save results
        custompath (str): local path from ~ to save results
    """

    def __init__(self, fitparams=None, inits=None, param_sets=None, basinparams=None, progress=True, custompath=None, nruns=10):

        self.fitparams = fitparams
        self.kind = fitparams.kind
        self.learn = fitparams.learn
        self.depends_on = fitparams.depends_on
        self.ssdMethod = fitparams.ssd_method

        if inits is None:
            inits = theta.get_default_inits(kind=self.kind, depends_on=self.depends_on, learn=self.learn)
        self.inits = deepcopy(inits)

        self.progress = progress
        self.callback = None
        self.constants = {}
        self.param_sets = param_sets

        self.set_basinparams()
        self.make_progress_bars(inits=True, basin=True)
        self.nruns=nruns

        if self.learn:
            data = self.get_trials_data()
            self.simRL = models_rl.Simulator(self.inits, data=data, fitparams=self.fitparams, ssdMethod=self.ssdMethod, constants=list(constants))
        self.sim = models.Simulator(self.inits, fitparams=self.fitparams, ssdMethod=self.ssdMethod)
        self.update()
        self.make_results_dir(custompath=custompath)


    def update(self, **kwargs):
        kw_keys = list(kwargs)
        kw = pd.Series(kwargs)

        if 'fitparams' in kw_keys:
            self.fitparams = kw.fitparams
        if 'basinparams' in kw_keys:
            self.basinparams = kw.basinparams

        self.ix = self.fitparams['ix']
        self.nlevels = self.fitparams['nlevels']
        self.learn = self.fitparams['learn']
        self.pcmap = self.fitparams.pcmap
        self.inits = self.fitparams.inits
        if 'nruns' in kw_keys:
            self.nruns = kw.nruns
        if self.learn:
            constants = self.simRL.constants
            if 'constants' in kw_keys:
                constants = kw['constants']
            data = self.get_trials_data()
            self.simRL.update(inits=self.inits, fitparams=self.fitparams, data=data, constants=constants, nruns=self.nruns)
        self.sim.update(inits=self.inits, fitparams=self.fitparams)


    def hop_around(self, param_sets=None, learn=False, ratesOnly=True, fitDynamics=True, rateParams=['B', 'C', 'R']):
        """ initialize model with niter randomly generated parameter sets
        and perform global minimization using basinhopping algorithm
        ::Arguments::
            p (dict):
                parameter dictionary
        ::Returns::
            parameter set with the best fit
        """
        if param_sets is None:
            self.sample_param_sets(learn)
            param_sets = self.param_sets
        self.make_progress_bars(inits=True, basin=True)
        self.xpopt, self.xfmin = [], []
        for i, p in enumerate(param_sets):
            self.update(inits=p, force='flat', learn=learn)
            if self.progress:
                self.ibar.update(value=i, status=i+1)
            popt, fmin = self.run_basinhopping(p=p, learn=learn, ratesOnly=ratesOnly, fitDynamics=fitDynamics, rateParams=rateParams, resetProgress=False)
            self.xpopt.append(popt)
            self.xfmin.append(fmin)
        if self.progress:
            self.gbar.clear()
            self.ibar.clear()
        popt = self.xpopt[np.argmin(self.xfmin)]
        if learn:
            self.rlpopt = deepcopy(popt)
        else:
            self.popt = deepcopy(popt)
        return popt


    def popt_array_to_dict(self, popt_arr, learn=False):
        if learn:
            pdict = dict(zip(self.simRL.pvary, popt_arr))
            pdict.update(self.simRL.fixedParams)
        elif self.sim.nlevels>1:
            nvary = self.sim.nvary
            pvary = self.sim.pvary
            pdict = {}; start = 0
            for i, n in enumerate(nvary):
                pdict[pvary[i]] = np.array(popt_arr[start:start+n])
                start = n
        else:
            pdict = dict(zip(self.sim.pvary, popt_arr))
        return pdict


    def run_basinhopping(self, p, learn=False, fitDynamics=True, ratesOnly=True, rateParams=['B', 'C', 'R'], resetProgress=True, return_all=False):
        bp = self.basinparams
        popt = deepcopy(p)

        if learn:
            if fitDynamics:
                costfx = self.simRL.cost_fx_rl
            else:
                costfx = self.simRL.cost_fx
            if ratesOnly:
                pkeys = [k for k in list(p) if k not in rateParams]
            else:
                pkeys = []
            self.simRL.update(inits=p, constants=pkeys)
            x0 = self.simRL.preproc_params(p, asarray=True)
        else:
            self.sim.set_pconstant_values_matrix(p)
            x0 = self.sim.pdict_to_array(p)
            costfx = self.sim.cost_fx

        if resetProgress:
            self.make_progress_bars(inits=True, basin=True)
        self.callback = self.gbar.reset(get_call=True, gbasin=resetProgress)

        # make list of init values for all pkeys included in fit
        self.set_basin_options(learn)
        # run basinhopping on simulator.basinhopping_minimizer func
        out = basinhopping(costfx, x0=x0, minimizer_kwargs=self.tnc_options, T=bp['T'], stepsize=bp['stepsize'], niter_success=bp['nsuccess'], niter=bp['niter'], interval=bp['interval'], take_step=self.custom_step, accept_test=self.accept_step, callback=self.callback)

        self.bh_results = out
        self.fit_info = out.lowest_optimization_result

        if return_all:
            return self.fit_info

        pdict = self.popt_array_to_dict(self.fit_info.x,  learn=learn)
        popt.update(pdict)
        fmin = self.fit_info.fun
        return popt, fmin


    def gradient_descent(self, p, learn=False):

        self.update(inits=p)
        fp = self.fitparams
        flat=False
        if fp.nlevels==1:
            flat=True

        optkws = {'xtol': fp['tol'], 'ftol': fp['tol'], 'maxfev': fp['maxfev'], 'maxiter': fp['maxiter']}

        if learn:
            sim = self.simRL
            pflat = sim.pflat
            pvary = sim.pvary
            x0 = sim.preproc_params(p, asarray=True)
            lmParams = theta.loadParameters_RL(inits=p, pflat=pflat, pvary=pvary, kind=self.kind)
        else:
            sim = self.sim
            sim.set_pconstant_values_matrix(p)
            # make lmfit Parameters object
            lmParams = theta.loadParameters(inits=p, pcmap=self.pcmap, is_flat=flat, kind=self.kind)
            lmParamsNames = list(lmParams.valuesdict())
            sim.update(lmParamsNames=lmParamsNames)
            self.lmParams = lmParams

        self.lbar = utils.GradientCallback(n=fp['maxfev'], fmin=10000)

        self.lmMin = minimize(sim.cost_fx_lmfit, lmParams, method=fp['method'], tol=fp['tol'],  options=optkws, iter_cb=self.lbar.callback)

        if hasattr(self, 'lbar'):
            self.lbar.clear()
        self.param_report = fit_report(self.lmMin.params)
        return self.assess_fit(flat=flat, learn=learn)


    def set_basinparams(self, **kwargs):
        """ dictionary of global fit parameters, passed to Optimizer/Simulator objects
        """
        if not hasattr(self, 'basinparams'):
            self.basinparams =  {'ninits': 3,
                                'nsamples': 2000,
                                'interval': 10,
                                'T': .025,
                                'stepsize': .01,
                                'niter': 400,
                                'nsuccess': 100,
                                'tol': 1.e-40,
                                'method': 'TNC',
                                'init_sample_method': 'best',
                                'progress': True,
                                'disp': True}
        else:
            # fill with kwargs for the upcoming fit
            for kw_arg, kw_val in kwargs.items():
                self.basinparams[kw_arg] = kw_val


    def set_basin_options(self, learn=False):
        if learn:
            sim = self.simRL
        else:
            sim = self.sim
        bp = self.basinparams
        xmin, xmax = format_basinhopping_bounds(sim.pvary, sim.nvary, self.kind)
        tnc_bounds = format_local_bounds(xmin, xmax)

        # define custom take_step and accept_test functions
        self.accept_step = BasinBounds(xmin, xmax)
        self.custom_step = HopStep(sim.pvary, nlevels=sim.nvary, stepsize=bp['stepsize'])
        self.tnc_options = {"method": 'TNC', 'bounds': tnc_bounds, 'tol': bp['tol'], 'options': {'xtol': bp['tol'], 'ftol': bp['tol']}}

        if self.progress:
            if not hasattr(self, 'gbar') or self.gbar.pbar.displayed==False:
                self.make_progress_bars(inits=False, basin=True)
            self.callback = self.gbar.reset(get_call=True, gbasin=False, history=False)


    def sample_param_sets(self, learn=False, fitDynamics=True):
        if learn:
            self.sample_rl_theta(fitDynamics=fitDynamics)
        else:
            self.sample_theta()


    def sample_theta(self):
        """ sample *nsamples* (default=5000, see set_fitparams) different
        parameter sets (param_sets) and get model yhat for each set (param_yhats)
        """
        pkeys = self.sim.theta.keys().tolist()
        nsamples = self.basinparams['nsamples']
        if not hasattr(self, 'Theta'):
            Params = theta.random_inits(pkeys, ninits=nsamples, kind=self.kind, as_list=True)
            yhats = np.vstack([self.sim.simulate_model(p) for p in Params])
            yhatsDF = pd.DataFrame(yhats)
            self.Theta = Params
            self.ThetaYhat = yhatsDF.copy()
        Params = self.Theta
        yhatsDF = self.ThetaYhat.copy()
        psets = self.filter_params(Params, yhatsDF)
        self.param_sets = psets


    def sample_rl_theta(self, fitDynamics=True):
        if fitDynamics:
            costfx = self.simRL.cost_fx_rl
        else:
            costfx = self.simRL.cost_fx
        nsamples = self.basinparams['nsamples']
        nkeep = self.basinparams['ninits']
        idxParams = pd.DataFrame([pd.Series(self.popt)]*nsamples)
        rlParams = theta.random_inits(['B', 'C', 'R'], ninits=nsamples, kind=self.kind, as_list=True)
        rlParamsDF = pd.concat([idxParams, pd.DataFrame(rlParams.tolist())], axis=1)
        rlParams = np.array(rlParamsDF.to_dict('records'))

        self.simRL.update(inits=self.popt, constants=list(self.popt))
        cost = np.vstack([costfx(p) for p in rlParams])
        rlPsets = rlParams[np.argsort(np.hstack(cost))[:nkeep]]
        self.param_sets = rlPsets


    def filter_params(self, psets=None, yhatDF=None):
        """ sample *nsamples* (default=5000, see set_fitparams) different
        parameter sets (param_sets) and get model yhat for each set (param_yhats)
            if fit_on==subjects flat_y shape is (n_idx X ndata)
            elseif fit_on==average flat_y shape is (1 X ndata)
        """
        if psets is None:
            psets = self.Theta
            yhatDF = self.ThetaYhat.copy()

        nkeep = self.basinparams['ninits']
        y = self.fitparams.y
        wts = self.fitparams.wts
        keys = yhatDF.columns.tolist()
        ySeries = pd.Series(dict(zip(keys, y)))
        wSeries = pd.Series(dict(zip(keys, wts)))
        psets = np.asarray(psets)
        diff = yhatDF - ySeries
        wDiff = diff * wSeries
        sqDiff = wDiff.apply(lambda x: x**2, axis=0)
        sseDF = sqDiff.apply(lambda x: np.sum(x), axis=1)
        sseVals =sseDF.values
        bestIX = sseVals.argsort()[:nkeep]
        return psets[bestIX]


    def get_trials_data(self):
        if self.fitparams.fit_on=='subjects':
            idx = np.int(self.fitparams['idx'])
            return self.data[self.data.idx==idx]
        else:
            return self.data.copy()


    def assess_fit(self, flat=True, learn=False):
        """ wrapper for analyze.assess_fit calculates and stores
        rchi, AIC, BIC and other fit statistics
        ::Arguments::
            flat (bool):
                if flat, yhat have ndim=1, else ndim>1
        ::Returns::
            yhat (array), finfo (pd.Series), popt (dict)
            see gradient_descent() docstrings
        """
        fp = deepcopy(self.fitparams)
        if learn:
            y = np.hstack([self.simRL.rtBlocks, self.simRL.saccBlocks])
            wts = np.hstack([self.simRL.rt_weights, self.simRL.sacc_weights])
        else:
            y = self.sim.y.flatten()
            wts = self.sim.wts.flatten()
        # gen dict of lmfit optimized Parameters object
        popt = dict(self.lmMin.params.valuesdict())

        # TODO: extract, calculate, and store std.errors of popt
        # presults is scipy.minimize object (how to get from lmParams?)
        # presults = self.lmMin.params (maybe?)
        # then take sqrt of the diag. of the hessian to get errors
        # poptErr = np.sqrt(np.diag(presults.hess_inv.todense()))

        if not self.learn:
            # un-vectorize all parameters except conditionals
            popt = theta.scalarize_params(popt, self.pcmap)
            if fp.nlevels>1:
                popt = theta.pvary_levels_to_array(popt, self.pcmap)
        finfo = pd.Series()
        # get model-predicted yhat vector
        #self.sim.simulate_model(popt)#(self.lmMin.residual / wts) + y
        fp['yhat'] = (self.lmMin.residual / wts) + y
        # fill finfo dict with goodness-of-fit info
        finfo['idx'] = fp.idx
        finfo['pvary'] = '_'.join(list(fp.depends_on))
        finfo['cnvrg'] = self.lmMin.success
        finfo['nfev'] = self.lmMin.nfev
        finfo['nvary'] = len(self.lmMin.var_names)
        finfo['chi'] = np.sum((wts*(fp['yhat'] - y))**2)
        finfo['ndata'] = len(fp['yhat'])
        finfo['df'] = finfo.ndata - finfo.nvary
        finfo['rchi'] = finfo.chi / finfo.df
        finfo['logp'] = finfo.ndata * np.log(finfo.chi / finfo.ndata)
        try:
            finfo['AIC'] = self.lmMin.aic
            finfo['BIC'] = self.lmMin.bic
        except Exception:
            finfo['AIC'] = finfo.logp + 2 * finfo.nvary
            finfo['BIC'] = finfo.logp + np.log(finfo.ndata) * finfo.nvary
        return finfo, popt, fp['yhat']


    def write_results(self, finfo, popt, yhat):
        """ logs fit info to txtfile, fills yhatdf and fitdf
        """
        self.yhatdf = self.model.handler.fill_yhatdf(yhat=yhat, fitparams=self.fitparams)
        self.fitdf = self.model.handler.fill_fitdf(finfo=finfo, fitparams=self.fitparams)
        self.poptdf = self.model.handler.fill_poptdf(popt=popt, fitparams=self.fitparams)


    def save_results(self, custompath=None, saveobserved=False):
        """ Saves yhatdf and fitdf results to model output dir
        ::Arguments::
            saveobserved (bool):
                if True will write observedDF & wtsDF to
                model output dir
        """
        if custompath is not None:
            self.make_results_dir(custompath=custompath)
        fname = self.model.model_id
        if self.model.is_nested:
            fname='nested_models'
        resultsPath = os.path.join(self.resultsdir, fname)
        make_fname = lambda savestr: '_'.join([resultsPath, savestr+'.csv'])
        self.yhatdf.to_csv(make_fname('yhat'), index=False)
        self.fitdf.to_csv(make_fname('finfo'), index=False)
        self.poptdf.to_csv(make_fname('popt'), index=False)
        if saveobserved:
            self.model.observedDF.to_csv(make_fname('observed_data'))
            self.model.wtsDF.to_csv(make_fname('cost_weights'))


    def log_fit_info(self, finfo=None, popt=None, yhat=None):
        """ write meta-information about latest fit
        to logfile (.txt) in working directory
        """
        fp = deepcopy(self.fitparams)
        fp['yhat'] = yhat
        # lmfit-structured fit_report to write in log file
        param_report = self.param_report
        # log all fit and meta information in working directory
        messages.logger(param_report, finfo=finfo, popt=popt, fitparams=fp, kind=fp.kind, fit_on=fp.fit_on, savepath=self.resultsdir)


    def make_results_dir(self, custompath=None):
        savedir = os.path.expanduser('~')
        if custompath is not None:
            savedir = os.path.join(savedir, custompath)
            if not os.path.isdir(savedir):
                os.mkdir(savedir)
        self.resultsdir = savedir


    def make_progress_bars(self, inits=True, basin=True, lBasin=True):
        bp = self.basinparams
        if inits:
            ninits = bp['ninits']
            status=' / '.join(['Inits {}', '{}'.format(ninits)])
            self.ibar = utils.PBinJ(n=ninits, color='b', status=status)
        if basin:
            if hasattr(self, 'gbar'):
                self.gbar.clear()
            #fmin = self.sim.cost_fx(self.param_sets[0][0])
            self.gbar = utils.BasinCallback(n=bp['nsuccess'], fmin=10000)
            self.callback = self.gbar.reset(get_call=True)
        if lBasin:
            fp = self.fitparams
            if hasattr(self, 'lbar'):
                self.lbar.clear()
            self.lbar = utils.GradientCallback(n=fp['maxfev'], fmin=10000)
            self.callback = self.lbar.reset(get_call=True)



#!usr/bin/env python
from __future__ import division
from future.utils import listvalues
import pandas as pd
import numpy as np
from copy import deepcopy
from numpy import array
from scipy.stats.distributions import norm, gamma, uniform
from lmfit import Parameters as lmParameters


def ezdiff(rt, correct, s=1.):
    """ estimate ezDDM parameters
    """
    logit = lambda p:np.log(p/(1-p))
    pc = np.mean(correct)

    # subtract or add 1/2 an error to prevent division by zero
    if pc==1.0:
        pc=1 - 1/(2*len(correct))
    if pc==0.5:
        pc=0.5 + 1/(2*len(correct))
    MRT = np.mean(rt[correct==1])
    VRT = np.var(rt[correct==1])

    r = (logit(pc)*(((pc**2) * logit(pc)) - pc*logit(pc) + pc - 0.5))/VRT
    v = np.sign(pc-0.5)*s*(r)**0.25
    a = (s**2 * logit(pc))/v
    y = (-1*v*a)/(s**2)
    MDT = (a/(2*v))*((1-np.exp(y))/(1+np.exp(y)))
    tr = MRT-MDT
    return([a, v, tr])


def filter_params(psets, yhatdf, ySeries, wSeries, nkeep=3):
    """ sample *nsamples* (default=5000, see set_fitparams) different
    parameter sets (param_sets) and get model yhat for each set (param_yhats)
        if fit_on==subjects flat_y shape is (n_idx X ndata)
        elseif fit_on==average flat_y shape is (1 X ndata)
    """
    psets = np.asarray(psets)
    diff = yhatdf - ySeries
    wDiff = diff * wSeries
    sqDiff = wDiff.apply(lambda x: x**2, axis=0)
    sseDF = sqDiff.apply(lambda x: np.sum(x), axis=1)
    sseVals =sseDF.values
    bestIX = sseVals.argsort()[:nkeep]
    return psets[bestIX]


def random_inits(pkeys, ninits=1, kind='dpm', mu=None, sigma=None, as_list=False, multi=False, force_normal=False):
    """ random parameter values for initiating model across range of
    parameter values.
    ::Arguments::
        pkeys (list/dict):
            list of parameter names to sample, makes list of dict keys
            if pkeys is dictionary type
    ::Returns::
        params (dict):
            dictionary with pkeys as keys with values as 1d arrays ninits long
            of randomly sampled parameter values
    """
    if isinstance(pkeys, dict):
        pkeys = list(pkeys)
    params = {}
    for pk in pkeys:
        params[pk] = init_distributions(pk, nrvs=ninits, kind=kind, mu=mu, sigma=sigma, multi=multi, force_normal=force_normal)
    if multi:
        # vi_perc ~ U(.05, .95) --> vi = vi_perc*vd
        params['vi'] = params['vi']*params['vd']
    if as_list:
        params = np.array([{pk: params[pk][i] for pk in pkeys} for i in range(ninits)])
    return params


def loadParameters(inits=None, is_flat=False, kind=None, pcmap={}):
    """ Generates and returns an lmfit Parameters object with
    bounded parameters initialized for flat or non flat model fit
    """
    lmParams = lmParameters()
    pnames = ['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso', 'C', 'B', 'R']
    pfit = list(set(inits.keys()).intersection(pnames))
    bounds = get_bounds(kind=kind)

    for pkey, pclist in pcmap.items():
        if is_flat:
            break  # exit
        pfit.remove(pkey)
        if hasattr(inits[pkey], '__iter__'):
            vals = inits[pkey]
        else:
            vals = inits[pkey] * np.ones(len(pclist))
        mn = bounds[pkey][0]
        mx = bounds[pkey][1]
        for k, v in zip(pclist, vals):
            if isinstance(v, np.ndarray):
                v = np.asscalar(v)
            lmParams.add(k, value=v, vary=True, min=mn, max=mx)
    for pk in pfit:
        inits = scalarize_params(inits, pfit)
        if is_flat:
            mn = bounds[pk][0]
            mx = bounds[pk][1]
            lmParams.add(pk, value=inits[pk], vary=True, min=mn, max=mx)
        elif isinstance(inits[pk], np.ndarray):
            pk_clist = ['_'.join([pk, pcond.split('_')[1]]) for pcond in pclist]
            for pname, pvals in zip(pk_clist, inits[pk]):
                lmParams.add(pname, value=pvals, vary=False)
        else:
            lmParams.add(pk, value=inits[pk], vary=False)
    return lmParams


def loadParameters_RL(inits, pvary, pflat, nlevels=1, kind='xdpm'):
    """ Generates and returns an lmfit Parameters object with
    bounded parameters initialized for flat or non flat model fit
    """
    lmParams = lmParameters()
    bounds = get_bounds(kind=kind)
    if nlevels > 1:
        varyIDs = ['{}_{}'.format(pkey, str(l)) for l in np.arange(nlevels)]
    for pkey in pvary:
        mn = bounds[pkey][0]
        mx = bounds[pkey][1]
        for i in range(nlevels):
            varyID = pkey
            if nlevels>1:
                varyID = '{}_{}'.format(pkey, str(l))
            lmParams.add(varyID, value=inits[pkey], vary=True, min=mn, max=mx)
    for pk in pflat:
        lmParams.add(pk, value=inits[pk], vary=False)
    return lmParams


def clean_popt(params, pcmap):
    params = deepcopy(params)
    pvary = np.sort(list(pcmap))
    for i, p in enumerate(pvary):
        p_lvls = np.sort(pcmap[p])
        params[p] = array([params[p_lvl] for p_lvl in p_lvls])
    for p in list(params):
        if p in pvary.tolist():
            continue
        pval = np.mean(params[p])
        params[p] = pval * np.ones(p_lvls.size)
    return params


def scalarize_params(params, pcmap={}):
    """ scalarize all parameters in params dict """
    params = deepcopy(params)
    pvary = np.sort(list(pcmap)).tolist()
    if isinstance(params, pd.Series):
        params = params.to_dict()
    for p in list(params):
        if p in pvary:
            continue
        if hasattr(params[p], '__iter__'):
            try:
                params[p] = np.asscalar(params[p])
            except ValueError:
                params[p] = np.mean(params[p])
    return params


def pvary_levels_to_array(popt, pcmap={}):
    """ store optimized values of parameter in popt dict as array
        Example:
            from:   {'v_bsl': 1.15, 'v_pnl':1.10, ...}
            to:     {'v': array([1.15, 1.10]), ...}
    """
    for pname, pconds in pcmap.items():
        popt[pname] = np.array([popt[c] for c in pconds])
    return popt


def init_distributions(pkey, kind='dpm', mu=None, sigma=None, nrvs=25, tb=.65, multi=False, force_normal=False):
    """ sample random parameter sets to explore global minima (called by
    Optimizer method __hop_around__())
    """

    if mu is None:
        mu = {'a': .3, 'tr': .1, 'v': 5, 'ssv': -5, 'z': .1, 'xb': .01, 'sso': .15, 'vi': .35, 'vd': .5, 'C': .001, 'B':.1, 'R': .0005, 'si':.001}
    if sigma is None:
        sigma = {'a': .6, 'tr': .5, 'v': 2.5, 'ssv': 2.5, 'z': .05, 'xb': 1.99, 'sso': .2, 'vi': .4, 'vd': .5, 'C':.08, 'B':.4, 'R': .008, 'si':.14}
    if force_normal:
        normal_params = [pkey]
    else:
        normal_params = ['v', 'vd', 'ssv', 'z', 'Beta']
        gamma_params = ['a', 'tr', 'sso']
        uniform_params = ['vi', 'xb', 'C', 'B', 'R', 'si']

    if 'race' in kind:
        sigma['ssv'] = abs(mu['ssv'])
    if multi:
        bounds = get_multi_bounds(kind=kind)[pkey]
    else:
        bounds = get_bounds(kind=kind)[pkey]
    loc = mu[pkey]
    scale = sigma[pkey]

    # init and freeze dist shape
    if pkey in normal_params:
        dist = norm(loc, scale)
    elif pkey in gamma_params:
        dist = gamma(1.0, loc, scale)
    elif pkey in uniform_params:
        dist = uniform(loc, scale)

    # generate random variates
    rvinits = dist.rvs(nrvs)

    while rvinits.min() < bounds[0]:
        # apply lower limit
        ix = rvinits.argmin()
        rvinits[ix] = dist.rvs()
    while rvinits.max() > bounds[1]:
        # apply upper limit
        ix = rvinits.argmax()
        rvinits[ix] = dist.rvs()
    if pkey =='tr':
        rvinits = np.abs(rvinits)
    rvinits[rvinits<bounds[0]] = bounds[0]
    rvinits[rvinits>bounds[1]] = bounds[1]
    return rvinits


def get_bounds(kind='dpm', a=(.1, 1.1), tr=(.1, .6), v=(1., 10.), z=(.01, .9), ssv=(-10., -1.), xb=(.1, 2.5), si=(.001, .15), sso=(.05, .35), vd=(.1, 2.1), vi=(.1, 1.), Beta = (0.5, 5.), R=(.0001, .008), B=(.1, .4), C=(.001, .08)):
    """ set and return boundaries to limit search space
    of parameter optimization in <optimize_theta>
    """
    if 'irace' in kind:
        ssv = (abs(ssv[1]), abs(ssv[0]))
    bounds = {'a': a, 'tr': tr, 'v': v, 'ssv': ssv, 'vd':vd, 'vi':vi, 'z': z,
        'xb': xb, 'si': si, 'sso': sso, 'C': C, 'B': B, 'R': R, 'Beta': Beta}
    return bounds


def get_multi_bounds(kind='dpm', a=(.01, .7), tr=(.005, .5), v=(.1, 2.), z=(.01, .9), ssv=(-2.5, -.1), xb=(.1, 2.5), si=(.001, .08), sso=(.005, .3), vd=(.1, 2.1), vi=(.4, .8), R=(.0001, .008), B=(.1, .4), C=(.001, .08), Beta=(0.5, 5.)):
    """ set and return boundaries to limit search space
    of parameter optimization in <optimize_theta>
    """
    if 'irace' in kind:
        ssv = (abs(ssv[1]), abs(ssv[0]))
    bounds = {'a': a, 'tr': tr, 'v': v, 'ssv': ssv, 'vd':vd, 'vi':vi, 'z': z,
        'xb': xb, 'si': si, 'sso': sso, 'C': C, 'B': B, 'R': R, 'Beta': Beta}
    return bounds


def format_local_bounds(xmin, xmax):
    tupler = lambda xlim: tuple([xlim[0], xlim[1]])
    # return map((tupler), zip(xmin, xmax))
    return [tupler(xl) for xl in zip(xmin, xmax)]


def format_basinhopping_bounds(basin_keys, nlevels=1, kind='dpm'):
    """ returns separate lists of all parameter
    min and max values """
    allbounds = get_bounds(kind=kind)
    xmin, xmax = [], []
    for pk in basin_keys:
        xmin.append([allbounds[pk][0]] * nlevels)
        xmax.append([allbounds[pk][1]] * nlevels)
    xmin = np.hstack(xmin).tolist()
    xmax = np.hstack(xmax).tolist()
    return xmin, xmax


def get_stepsize_scalars(keys, nlevels=1):
    """ returns an array of scalars used by fit.HopStep() object
    to control stepsize of basinhopping algorithm for each parameter """
    # scalar_dict = {'a': .2, 'tr': .1, 'v': 1.1, 'vi': 1.1, 'vd': 1.1,
    #                'ssv': 1.1, 'z': .1, 'xb': 1.1, 'sso': .1, 'C': .001, 'B': .1, 'R':.1}
    # scalar_dict = {'a': .005, 'tr': .005, 'v': .01, 'vi': .01, 'vd': .01, 'ssv': .01, 'z': .001, 'xb': .1, 'sso': .001, 'C': .0002, 'B': .01, 'R':.01}
    # scalar_dict = {'a': .5, 'tr': .2, 'v': 1.5, 'vi': 1.1, 'vd': 1.5, 'ssv': 1.5,
    #     'z': .1, 'xb': 1.5, 'sso': .1, 'si': .01, 'C': .1, 'B': .2, 'R':.2, 'Beta': .2}
    # scalar_dict = {'a': .5, 'tr': .1, 'v': 1.5, 'vi': 1.5, 'vd': 1.5, 'ssv': 1.5,
    #         'z': .5, 'xb': 1.5, 'sso': .1, 'si': .1, 'C': .5, 'B': .5, 'R':.5, 'Beta': .2}
    scalar_dict = {'a': .3, 'tr': .5, 'v': 1., 'vi': 1., 'vd': 1.5,
        'ssv': 1., 'z': .5, 'xb': 1., 'sso': .1, 'si': .1, 'C': .5,
        'B': .5, 'R':.1, 'Beta': .2}
    nl_ones = np.ones(nlevels)
    stepsize_scalars = np.hstack([scalar_dict[k]*nl_ones for k in keys])
    if nlevels>1:
        stepsize_scalars = stepsize_scalars.squeeze()
    return stepsize_scalars


def get_default_inits(kind='dpm', depends_on={}, learn=False, ssdelay=False):
    """ if user does not provide inits dict when initializing Model instance,
    grab default dictionary of init params reasonably suited for Model kind
    """
    inits = {'a': 0.42, 'v': 1.25, 'tr': 0.2}
    pdep = list(depends_on)

    if 'dpm' in kind:
        inits['ssv'] = -1.3
    elif 'race' in kind:
        inits['ssv'] = 1.3
    if 'x' in kind and 'xb' not in list(inits):
        inits['xb'] = 1.5
    if 'si' in pdep:
        inits['si'] = .01
    if ssdelay:
        inits['sso'] = .08
    if learn:
        inits['C'] = .02
        inits['B'] = .15
        inits['R'] = .0015
    return inits


def check_inits(inits={}, depends_on={}, kind='dpm'):
    """ ensure inits dict is appropriate for Model kind
    """
    pdep = list(depends_on)
    if 'race' in kind or 'iact' in kind:
        inits['ssv'] = abs(inits['ssv'])
    elif 'dpm' in kind:
        inits['ssv'] = -abs(inits['ssv'])
    if 'pro' in kind:
        if 'ssv' in inits.keys():
            ssv = inits.pop('ssv')
    if 'x' in kind and 'xb' not in inits.keys():
        inits['xb'] = 1.5
    if 'si' in pdep and 'si' not in inits.keys():
        inits['si'] = .01
    if 'dpm' not in kind and 'z' in inits.keys():
        discard = inits.pop('z')
    if 'x' not in kind and 'xb' in inits:
        inits.pop('xb')
    # make sure inits only contains subsets of these params
    pnames = ['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso', 'C', 'B', 'R']
    pfit = list(set(list(inits)).intersection(pnames))
    return {pk: inits[pk] for pk in pfit}


def update_params(theta):
    if 't_hi' in theta.keys():
        theta['tr'] = theta['t_lo'] + np.random.uniform() * (theta['t_hi'] - theta['t_lo'])
    if 'z_hi' in theta.keys():
        theta['z'] = theta['z_lo'] + np.random.uniform() * (theta['z_hi'] - theta['z_lo'])
    if 'sv' in theta.keys():
        theta['v'] = theta['sv'] * np.random.randn() + theta['v']
    return theta


def get_intervar_ranges(theta):
    """ theta (dict): parameters dictionary
    """
    if 'st' in theta.keys():
        theta['t_lo'] = theta['tr'] - theta['st'] / 2
        theta['t_hi'] = theta['tr'] + theta['st'] / 2
    if 'sz' in theta.keys():
        theta['z_lo'] = theta['z'] - theta['sz'] / 2
        theta['z_hi'] = theta['z'] + theta['sz'] / 2
    return theta



#!usr/bin/env python
from __future__ import division
import pandas as pd
import numpy as np
import os
from numpy import array
from future.utils import listvalues
from sklearn.neighbors import KernelDensity
from scipy.stats.mstats import mquantiles as mq
from scipy.stats.mstats_extras import mjci
from scipy import optimize
import functools
from scipy.interpolate import interp1d
from scipy import stats
from scipy.optimize import leastsq, minimize, curve_fit


def blockify_trials(data, nblocks=5, conds=None, groups=['idx']):
    datadf = data.copy()
    if conds is not None:
        if type(conds) is str:
            conds = [conds]
        groups = groups + conds
    idxdflist = []
    for dfinfo, idxdf in datadf.groupby(groups):
        ixblocks = np.array_split(idxdf.trial.values, nblocks)
        blocks = np.hstack([[i+1]*arr.size for i, arr in enumerate(ixblocks)])
        idxdf = idxdf.copy()
        colname = 'block{}'.format(nblocks)
        idxdf[colname] = blocks
        idxdflist.append(idxdf)
    return pd.concat(idxdflist)


def rangl_data(data, ssd_method='all', quantiles=np.linspace(.01,.99,15), fit_on='average'):
    """ called by DataHandler.__make_dataframes__ to generate
    observed data arrays
    """
    data = data.copy()
    if 1 in data.ttype.unique():
        goQuery = 'ttype==1.'
        stopQuery = 'ttype==0.'
    else:
        goQuery = 'ttype=="go"'
        stopQuery = 'ttype=="stop"'
    gac = data.query(goQuery).acc.mean()
    grt = data.query('response==1 & acc==1').rt.values
    ert = data.query('response==1 & acc==0').rt.values
    gq = mq(grt, prob=quantiles)
    eq = mq(ert, prob=quantiles)
    data_vector = [gac, gq, eq]
    if 'ssd' in data.columns:
        stopdf = data.query(stopQuery)
        if 'probe' in stopdf.columns:
            stopdf = stopdf[stopdf.probe==1]
        if ssd_method=='all':
            sacc=stopdf.groupby('ssd').mean()['acc'].values
        elif ssd_method=='central':
            sacc = np.array([stopdf.mean()['acc']])
        data_vector.insert(1, sacc)
    return np.hstack(data_vector)
    

def calculate_qpdata(data, quantiles=np.linspace(.02,.98,10)):
    data = data[data.response==1].copy()
    cor = data[data['acc']==1]
    err = data[data['acc']==0]
    idx_qc = pd.DataFrame([mq(idxdf.rt.values, quantiles) for idx, idxdf in cor.groupby('idx')])
    idx_qe = pd.DataFrame([mq(idxdf.rt.values, quantiles) for idx, idxdf in err.groupby('idx')])
    quantMeans = np.vstack = ([idx_qc.mean(axis=0).values, idx_qe.mean(axis=0).values])
    quantErr = np.vstack([idx_qc.sem(axis=0).values, idx_qe.sem(axis=0).values])*1.96
    return quantMeans, quantErr

def idx_quant_weights(data, conds, quantiles=np.linspace(.02,.98,10), max_wt=3, bwfactors=None):
    """ calculates weight vectors for reactive RT quantiles by
    first estimating the SEM of RT quantiles for corr. and err. responses.
    (using Maritz-Jarrett estimatation: scipy.stats.mstats_extras.mjci).
    Then representing these variances as ratios.
    e.g.
          QSEM = mjci(rtvectors)
          wts = median(QSEM)/QSEM
    """
    idx_mjci = lambda x: mjci(x.rt, prob=quantiles)

    if bwfactors is not None:
        if not isinstance(bwfactors, list):
            bwfactors = [bwfactors]
        conds = [c for c in conds if c not in bwfactors]

    # get all trials with response recorded
    nlevels = np.sum([data[c].unique().size for c in conds])
    godf = data.query('response==1')
    # sort by ttype first so go(acc==1) occurs before stop(acc==0)
    ttype_ordered_groups = np.hstack([conds, 'ttype', 'acc']).tolist()
    godf_grouped = godf.groupby(ttype_ordered_groups)
    # apply self.idx_mjci() to estimate quantile CI's
    quant_err = np.hstack(godf_grouped.apply(idx_mjci).values)
    # calculate subject median across all conditions quantiles and accuracy
    # this implicitly accounts for n_obs as the mjci estimate of sem will be lower
    # for conditions with greater number of observations (i.e., more acc==1 in godf)
    idx_median = np.nanmedian(quant_err)
    idx_qratio = idx_median / quant_err
    # set extreme values to max_wt arg val
    idx_qratio[idx_qratio >= max_wt] = max_wt
    return idx_qratio.reshape(nlevels, -1)

def idx_acc_weights(data, conds=['flat'], ssd_method='all'):
    """ count number of observed responses across levels, transform into ratios
    (counts_at_each_level / np.median(counts_at_each_level)) for weight
    subject-level p(response) values in cost function.
    ::Arguments::
        df (DataFrame): group-level dataframe
        var (str): column header for variable to count responses
        conds (list): depends_on.values()
    """
    df = data.copy()
    index=['idx']
    if not 'flat' in conds:
        index = index + conds
    if 'ssd' in df.columns:
        if ssd_method=='all':
            split_by = 'ssd'
            df = df[df.ttype=='stop'].copy()
            if 'probe' in df.columns:
                df = df[df.probe==1].copy()
        else:
            split_by = 'ttype'
    else:
        split_by = conds
        _ = index.remove(split_by)
    df['n'] = 1
    countdf = df.pivot_table('n', index=index, columns=split_by, aggfunc=np.sum)
    #idx_pwts = countdf.values / countdf.median(axis=1).values[:, None]
    idx_pwts = countdf.values / countdf.median(axis=0).values
    if ssd_method=='all':
        go_wts = np.ones(countdf.shape[0])
        idx_pwts = np.concatenate((go_wts[:,None], idx_pwts), axis=1)
    return idx_pwts

def assess_fit(y, wts, yhat, nvary=5):
    # residual = wts * (yhat - y)
    # fp['yhat'] = (residual / wts) + y
    # fill finfo dict with goodness-of-fit info
    y = y.flatten()
    wts = wts.flatten()
    yhat = yhat.flatten()
    finfo = pd.Series()
    finfo['nvary'] = nvary
    finfo['chi'] = np.sum((wts*(yhat - y))**2)
    finfo['ndata'] = len(yhat)
    finfo['df'] = finfo.ndata - finfo.nvary
    finfo['rchi'] = finfo.chi / finfo.df
    finfo['logp'] = finfo.ndata * np.log(finfo.rchi)
    finfo['AIC'] = finfo.logp + 2 * finfo.nvary
    finfo['BIC'] = finfo.logp + np.log(finfo.ndata * finfo.nvary)
    return finfo

def pandaify_results(gort, ssrt, tb=.7, bootstrap=False, bootinfo={'nsubjects':25, 'ntrials':1000, 'groups':['ssd']}, ssd=np.array([[.2, .25, .3, .35, .4]]), clmap=None):
    nl, nssd, nssPer = ssrt.shape
    nl, ntrials = gort.shape
    nss = nssd * nssPer
    dfColumns=['cond', 'ttype', 'ssd', 'response', 'acc', 'rt', 'ssrt', 'trial']
    levelNames = np.hstack(listvalues(clmap)).tolist()
    # bootinfo['groups'] = bootinfo['groups'] + list(clmap)
    dfIndex = np.arange(ntrials)
    dfList = []
    for i in range(nl):
        ert = gort[i, :nss].reshape(ssrt[i].shape)
        goTrialOutcomes = np.hstack(np.where(gort[i, nss:] < tb, 1, 0))
        # ssTrialOutcomes = np.hstack(np.hstack(np.where(ert <= ssrt[i], 1, 0)))
        ssTrialOutcomes = np.hstack(np.hstack(np.where(ssrt[i] <= ert, 0, 1)))
        delays = np.append(np.hstack([[delay]*nssPer for delay in ssd[i]]), [1000]*nss)
        response = np.append(ssTrialOutcomes, goTrialOutcomes)
        responseTime = gort[i]
        ssResponseTime = np.append(np.hstack(np.hstack(ssrt[i])), [np.nan]*nss)
        ttype = np.append(np.zeros(nss), np.ones(nss))
        acc = np.where(ttype == response, 1, 0)
        cond = [levelNames[i]]*ntrials
        trial = np.arange(1, ntrials+1)
        dfData = [cond, ttype, delays, response, acc, responseTime, ssResponseTime, trial]
        df = pd.DataFrame(dict(zip(dfColumns, dfData)), index=dfIndex)
        dfList.append(df[dfColumns])
    df = pd.concat(dfList)
    df.reset_index(drop=True, inplace=True)
    # stopdf = resultsdf[resultsdf.ttype==0.]
    df.loc[(df.rt==1000.)&(df.ssrt==1000.)&(df.ttype==0.), 'response'] = 0
    df.loc[(df.rt==1000.)&(df.ssrt==1000.)&(df.ttype==0.), 'acc'] = 1

    if bootstrap:
        resultsdf = bootstrap_data(resultsdf, nsubjects=bootinfo['nsubjects'], n=bootinfo['ntrials'], groups=bootinfo['groups'])
    return resultsdf

def bootstrap_data(data, nsubjects=25, n=120, groups=['cond', 'ssd']):
    """ generates n resampled datasets using rwr()
    for bootstrapping model fits
    """
    # groups=['ttype', 'ssd']
    df = data.copy()
    bootlist = []
    idxList = []
    if 'idx' in df.columns:
        df = df.drop('idx', axis=1)
    if n == None:
        n = len(df)
    # bootdfList = []
    # for level, levelDF in df.groupby(groups[0]):
    bootlist = []
    for idx in range(nsubjects):
        boots = df.reset_index(drop=True)
        orig_ix = np.asarray(boots.index[:])
        resampled_ix = rwr(orig_ix, get_index=True, n=n)
        bootdf = df.iloc[resampled_ix]
        bootdf = bootdf.copy()
        bootdf.insert(0, 'idx', idx)
        bootdf['trial'] = np.arange(bootdf.shape[0])
        bootlist.append(bootdf)
    # concatenate and return all resampled conditions
    bootdf = pd.concat(bootlist)
    bootdf.reset_index(drop=True, inplace=True)
    # bootdfList.append(bootdf)
    # bootdf = pd.concat(bootdfList)
    return bootdf

def rwr(X, get_index=False, n=None):
    """
    Modified from http://nbviewer.ipython.org/gist/aflaxman/6871948
    """
    if isinstance(X, pd.Series):
        X = X.copy()
        X.index = range(len(X.index))
    if n == None:
        n = len(X)
    resample_i = np.floor(np.random.rand(n) * len(X)).astype(int)
    X_resample = (X[resample_i])
    if get_index:
        return resample_i
    else:
        return X_resample

def calcPostErrAdjust(df, ntrials=200):
    if type(df.ttype.values[0]) is str:
        ssErrDF = df[(df.ttype=='stop')&(df.response==1)&(df.rt<.68)]
    else:
        ssErrDF = df[(df.ttype==0.)&(df.response==1)&(df.rt<.68)]
    ssErrRT = ssErrDF.rt.mean()
    adjustIX = ssErrDF.index.values + 1
    adjustIX[-1] = ssErrDF.index.values[-1]
    adjustDF = df.loc[adjustIX, :]
    PostErrRT = adjustDF[(adjustDF.response==1)&(adjustDF.rt<.68)].rt.mean()
    PostErrAdjust = PostErrRT - ssErrRT
    return PostErrAdjust

def calcTargetAdjust(df, ntrials=200):
    if type(df.ttype.values[0]) is str:
        GoDF = df[(df.ttype=='go')&(df.response==1)&(df.rt<.68)]
    else:
        GoDF = df[(df.ttype==1.)&(df.response==1)&(df.rt<.68)]
    GoRT = GoDF.rt.mean()
    adjustIX = GoDF.index.values + 1
    adjustIX[-1] = GoDF.index.values[-1]
    adjustDF = df.loc[adjustIX]
    GoAdjustRT = adjustDF[(adjustDF.response==1)&(adjustDF.rt<.68)].rt.mean()
    TargetAdjust =  GoAdjustRT - GoRT
    return TargetAdjust


def estimate_timeboundary(data):
    goDF = data[data.response==1].copy()
    return np.int(np.round(goDF.rt.max(), 2)*1000) * .001

def determine_ssd_method(stopdf):
    ssd_n = [df.size for _, df in stopdf.groupby('ssd')]
    # test if equal # of trials per ssd & return ssd_n
    all_equal_counts = ssd_n[1:] == ssd_n[:-1]
    if all_equal_counts:
        return 'all'
    else:
        return'central'

def get_model_ssds(stopdf, conds, ssd_method='all', scale=.001, bwfactors=None):

    if bwfactors is not None:
        if not isinstance(bwfactors, list):
            bwfactors = [bwfactors]
        conds = [c for c in conds if c not in bwfactors]

    if ssd_method == 'all':
        get_df_ssds = lambda df: df.groupby(conds).ssd.unique().values
        cond_ssds =  [get_df_ssds(df) for _,df in stopdf.groupby('idx')]

    elif ssd_method == 'central':
        mean_cond_ssd_df = stopdf.pivot_table('ssd', index='idx', columns=conds)
        cond_ssds = list(mean_cond_ssd_df.values)

    ssds = [np.sort(np.vstack(ssds))*scale for ssds in cond_ssds]
    groups = conds + ['idx']
    groupInfo = stopdf.groupby(groups).mean().reset_index()[groups[::-1]]
    ssdDF = pd.concat([groupInfo, pd.DataFrame(np.vstack(ssds).squeeze())], axis=1)

    return ssdDF


def load_nested_popt_dictionaries(fitdf, params={}, nlevels=1):
    mparams = list(params)
    params_dict = {}

    for mid in fitdf.idx.values:
        freeparams = mid.split('_')
        constants = [p for p in mparams if p not in freeparams]
        select_params = fitdf[fitdf.idx==mid][mparams]
        pdict = select_params.to_dict('records')[0]
        for p in constants:
            pdict[p] = pdict[p] * np.ones(nlevels)
        params_dict[mid] = pdict

    return params_dict


def fill_nan_vals(coldata):
    coldata = coldata.copy()
    coldata[coldata.isnull()] = np.nanmean(coldata.values)
    return coldata


def remove_outliers(data, sd=1.5, verbose=False):
    df = data.copy()
    ssdf = df[df.response == 0]
    godf = df[df.response == 1]
    bound = godf.rt.std() * sd
    rmslow = godf[godf['rt'] < (godf.rt.mean() + bound)]
    clean_go = rmslow[rmslow['rt'] > (godf.rt.mean() - bound)]
    clean = pd.concat([clean_go, ssdf])
    if verbose:
        pct_removed = len(clean) * 1. / len(df)
        print("len(df): %i\nbound: %s \nlen(cleaned): %i\npercent removed: %.5f" % (len(df), str(bound), len(clean), pct_removed))
    sortby = 'idx'
    if 'trial' in clean.columns:
        sortby = ['idx', 'trial']
    cleandf = clean.sort_values(by=sortby)
    return cleandf

def ensure_numerical_wts(wts, flat_wts):
    # test inf
    wts[np.isinf(wts)] = np.median(wts[~np.isinf(wts)])
    flat_wts[np.isinf(flat_wts)] = np.median(flat_wts[~np.isinf(flat_wts)])
    # test nan
    wts[np.isnan(wts)] = np.median(wts[~np.isnan(wts)])
    flat_wts[np.isnan(flat_wts)] = np.median(flat_wts[~np.isnan(flat_wts)])
    return wts, flat_wts

def kde_fit_quantiles(rtquants, nsamples=1000, bw=.1):
    """ takes quantile estimates and fits cumulative density function
    returns samples to pass to sns.kdeplot()
    """
    kdefit = KernelDensity(kernel='gaussian', bandwidth=bw).fit(rtquants)
    samples = kdefit.sample(n_samples=nsamples).flatten()
    return samples

def scurve_interpolate(x, y, kind='cubic'):
    interpol_fx = interp1d(x, y, kind=kind)
    xsim = np.linspace(x[0], x[-1], 50, endpoint=True)
    ysim = interpol_fx(xsim)
    return xsim, ysim

def scurve_poly_fit(x, y, n=20):
    polysim = lambda p, pi, x, xi: p[pi]*x**xi
    ix = np.arange(n+1)
    poly_ix = zip(ix, ix[::-1])
    p = np.polyfit(x,y,n)
    xsim = np.linspace(x.min()-.05, x.max()+.08, 2500, endpoint=True)
    ysim = np.sum([polysim(p, pi, xsim, xi) for pi, xi in poly_ix],axis=0)
    return xsim, ysim

def fit_sigmoid(x, y):
    #x, y = scurve_interpolate(x, y)
    initParams = [np.mean(x), .8, .5]
    results = minimize(sigmoid, initParams, args=(x, y), method='Nelder-Mead')
    popt = results.x
    xSim = np.linspace(x.min()-.01, x.max()+.01, 1000, endpoint=True)
    ySim = 1. / (1. + np.exp(popt[0]*(xSim-popt[1])))
    return xSim, ySim


def sigmoid(params, xdata, ydata):
    k = params[0]
    x0 = params[1]
    sd = params[2]
    yPred = 1. / (1. + np.exp(k*(xdata-x0)))
    # Calculate negative log likelihood
    LL = -np.sum(stats.norm.logpdf(ydata, loc=yPred, scale=sd ))
    return(LL)


def logistic4(x, a, b, c, d):
    """4PL lgoistic equation."""
    return ((a-d)/(1.0+((x/c)**b))) + d


def logistic_residuals(p, y, x):
    """Deviations of data from fitted 4PL curve"""
    a,b,c,d = p
    err = y-logistic4(x, a, b, c, d)
    return err


def logistic_peval(x, p):
    """Evaluated value at x with current parameters."""
    a,b,c,d = p
    return logistic4(x, a, b, c, d)


def fit_logistic(x, y):
    # Initial guess for parameters
    p0 = [0., .5, 1, 1]
    popt = leastsq(logistic_residuals, p0, args=(y, x))
    xSim = np.linspace(x.min()-.1, x.max()+.1, 1000, endpoint=True)
    ySim = logistic_peval(xSim, popt[0])
    return xSim, ySim


def residuals(p, x, y):
    return y - sigmoid(p, x)


def resize(arr, lower=0.0, upper=1.0):
    arr = arr.copy()
    if lower > upper:
        lower, upper = upper, lower
    arr -= arr.min()
    arr *= (upper - lower) / arr.max()
    arr += lower
    return arr

def get_observed_vector(rt, quantiles=array([10, 30, 50, 70, 90])):
    """ takes array of rt values and returns binned counts (trials
    that fall between each set of quantiles in quantiles). also returns
    the total number of observations (len(rt)) and the RT values at those
    quantiles (rtquant)
    """
    inter_quantiles = array([quantiles[0] - 0] + [quantiles[i] - quantiles[i - 1] for i in range(1, len(quantiles))] + [100 - quantiles[-1]])
    rtquant = mq(rt, prob=quantiles * .01)
    ocounts = np.ceil((inter_quantiles) * .01 * len(rt)).astype(int)
    n_obs = np.sum(ocounts)
    return [ocounts, rtquant, n_obs]

def get_expected_vector(simrt, obsinfo):
    """ calculates the expected frequencies of responses for a
    set of simulated RTs, given. obsinfo is output of
    get_observed_vector() -->  [ocounts, rtquant, n_obs]
    simrt = pd.Series(simrt)
    """
    counts, q, n_obs = obsinfo[0], obsinfo[1], obsinfo[2]
    first = array([len(simrt[simrt.between(simrt.min(), q[0])]) / len(simrt)]) * n_obs
    middle = array([len(simrt[simrt.between(q[i - 1], q[i])]) / len(simrt) for i in range(1, len(q))]) * n_obs
    last = array([len(simrt[simrt.between(q[-1], simrt.max())]) / len(simrt)]) * n_obs
    expected = np.ceil(np.hstack([first, middle, last]))
    return expected

def ssrt_calc(df, avgrt=.3):
    dfstp = df.query('ttype=="stop"')
    dfgo = df.query('choice=="go"')
    pGoErr = ([idf.response.mean() for ix, idf in dfstp.groupby('idx')])
    nlist = [int(pGoErr[i] * len(idf)) for i, (ix, idf) in enumerate(df.groupby('idx'))]
    GoRTs = ([idf.rt.sort(inplace=False).values for ix,idf in dfgo.groupby('idx')])
    ssrt_list = ([GoRTs[i][nlist[i]] for i in np.arange(len(nlist))]) - avgrt
    return ssrt_list

def get_obs_quant_counts(df, quantiles=([.10, .30, .50, .70, .90])):
    if type(df) == pd.Series:
        rt = df.copy()
    else:
        rt = df.rt.copy()
    inter_quantiles = [quantiles[0] - 0] + [quantiles[i] - quantiles[i - 1] for i in range(1, len(quantiles))] + [1.00 - quantiles[-1]]
    obs_quant = mq(rt, prob=quantiles)
    observed = np.ceil((inter_quantiles) * len(rt) * .94).astype(int)
    return observed, obs_quant

def get_exp_counts(simdf, obs_quant, n_obs, quantiles=([.10, .30, .50, .70, .90])):
    if type(simdf) == pd.Series:
        simrt = simdf.copy()
    else:
        simrt = simdf.rt.copy()
    exp_quant = mq(simrt, prob=quantiles)
    oq = obs_quant
    expected = np.ceil(np.diff([0] + [pscore(simrt, oq_rt) * .01 for oq_rt in oq] + [1]) * n_obs)
    return expected, exp_quant

def pwts_group_error_calc(handler):
    """ get stdev across subjects (and any conds) in observedDF
    weight perr by inverse of counts for each resp. probability measure
    previously a bound method of dfhandler --> DataHandler class
    """
    groupedDF = handler.observedDF.groupby(handler.conds)
    perr = groupedDF.agg(np.nanstd).loc[:, handler.p_cols].values
    counts = groupedDF.count().loc[:, handler.p_cols].values
    nsplits = handler.nlevels * handler.nconds
    ndata = len(handler.p_cols)
    # replace stdev of 0 with next smallest value in vector
    perr[perr==0.] = perr[perr>0.].min()
    p_wt_bycount = perr * (1./counts)
    # set wts equal to ratio --> median_perr / all_perr_values
    pwts_ratio = np.nanmedian(p_wt_bycount, axis=1)[:, None] / p_wt_bycount
    # set extreme values to max_wt arg val
    pwts_ratio[pwts_ratio >= handler.max_wt] = handler.max_wt
    # shape pwts_ratio to conform to wtsDF
    idx_pwts = np.array([pwts_ratio]*handler.nidx)
    return idx_pwts.reshape(handler.nidx * nsplits, ndata)

def get_intersection(iter1, iter2):
    """ get the intersection of two iterables ("items in-common")
    """
    intersect_set = set(iter1).intersection(set(iter2))
    return ([i for i in intersect_set])



#!/usr/local/bin/env python
from __future__ import division
from copy import deepcopy
import numpy as np
import pandas as pd
from numpy import array
from radd.models import Simulator
from radd.CORE import RADDCore
from radd import vis
from radd.tools import utils, analyze, messages
from radd.tools.analyze import pandaify_results, rangl_data


class Model(RADDCore):
    """ Main class for instantiating, fitting, and simulating models.
    Inherits from RADDCore parent class (see CORE module).
    ::Arguments::
        data (pandas DF):
            data frame with columns 'idx', 'rt', 'acc', 'ttype', 'response',
            <Condition Name> declared in depends_on values
        kind (str):
            declares model type ['dpm', 'irace', 'pro']
            append 'x' to front of model name to include a dynamic
            bias signal in model
        inits (dict):
            dictionary of parameters (v, a, tr, ssv, z) used to initialize model
        fit_on (str):
            set if model fits 'average', 'subjects', 'bootstrap' data
        depends_on (dict):
            set parameter dependencies on task conditions
            (ex. depends_on={'v': 'Condition'})
        weighted (bool):
            if True (default), perform fits using a weighted least-squares approach
        quantiles (array):
            set the RT quantiles used to fit model
    """

    def __init__(self, data=pd.DataFrame, kind='xdpm', inits=None, fit_on='average', depends_on={'all':'flat'}, weighted=True, ssd_method=None, learn=False, bwfactors=None, custompath=None, ssdelay=False, quantiles=np.array([.05,.1,.2,.3,.4,.5,.6,.7,.8,.9,.95])):

        super(Model, self).__init__(data=data, inits=inits, fit_on=fit_on, depends_on=depends_on, kind=kind, quantiles=quantiles, weighted=weighted, ssd_method=ssd_method, learn=learn, bwfactors=bwfactors, custompath=custompath, ssdelay=ssdelay)


    def optimize(self, flat_popt=None, plotfits=True, saveplot=False, saveresults=True, saveobserved=False, custompath=None, progress=False):
        """ Method to be used for accessing fitting methods in Optimizer class
        see Optimizer method optimize()
        ::Arguments::
            plotfits (bool):
                if True (default), plot model predictions over observed data
            saveplot (bool):
                if True (default is False), save plots to "~/<self.model_id>/"
            saveresults (bool):
                if True (default), save fitdf, yhatdf, and txt logs to "~/<self.model_id>/"
            saveobserved (bool):
                if True (default is False), save observedDF to "~/<self.model_id>/"
            custompath (str):
                path starting from any subdirectory of "~/" (e.g., home).
                all saved output will write to "~/<custompath>/<self.model_id>/"
            progress (bool):
                track progress across ninits and basinhopping
        """
        self.toggle_pbars(progress=progress)
        self.custompath=custompath

        for ix in range(len(self.observed)):
            if hasattr(self, 'idxbar'):
                self.idxbar.update(value=ix, status=self.idx[ix])
            if flat_popt is None:
                self.set_fitparams(ix=ix, force='flat', nlevels=1)
                self.sample_param_sets()
                flat_popt = self.optimize_flat(self.param_sets)
            if not self.is_flat:
                self.set_fitparams(ix=ix, force='cond')
                self.optimize_conditional(flat_popt)
            if plotfits:
                self.plot_model_fits(save=saveplot)
            if saveresults:
                self.handler.save_results(saveobserved)
        if progress and not self.is_nested:
            self.opt.gbar.clear()
            self.opt.ibar.clear()


    def optimize_flat(self, param_sets=None):
        """ optimizes flat model to data collapsing across all conditions
        ::Arguments::
            None
        ::Returns::
            None
        ::Attributes Created::
            yhat_flat (array): model-predicted data array (ndim = self.levels)
            finfo_flat (pd.Series): fit info (AIC, BIC, chi2, redchi, etc)
            popt_flat (dict): optimized parameters dictionary
        """
        # Global Optimization w/ Basinhopping (+TNC)
        self.basin = self.opt.hop_around(param_sets)
        # Flat Simplex Optimization of Parameters at Global Minimum
        self.finfo, self.popt, self.yhat = self.opt.gradient_descent(p=self.basin)
        if self.is_flat:
            self.write_results()
        return self.popt


    def optimize_conditional(self, flatp, hop=False):
        """ optimizes full model to all conditions in data
        ::Arguments::
            None
        ::Returns::
            None
        ::Attributes Created::
            yhat (array): model-predicted data array (ndim = self.levels)
            finfo (pd.Series): fit info (AIC, BIC, chi2, redchi, etc)
            popt (dict): optimized parameters dictionary
            flat_popt (dict): deepcopy of popt
        """
        basin = self.__check_inits__(deepcopy(flatp))
        self.set_fitparams(force='cond', inits=basin)
        # Pretune Conditional Parameters
        if hop:
            basin, fmin = self.opt.run_basinhopping(basin)
            self.opt.update(inits=basin)
        # Final Simplex Optimization
        self.finfo, self.popt, self.yhat = self.opt.gradient_descent(p=basin)
        # self.write_results()


    def nested_optimize(self, models=[], flatp=None, saveplot=True, plotfits=True, custompath=None, progress=False, saveresults=True, saveobserved=False):
        """ optimize a series of models using same init parameters where the i'th model
            has depends_on = {<models[i]> : <cond>}.
            NOTE: only for models with fit_on='average'
        ::Arguments::
            models (list):
                list of depends_on dictionaries to fit using a single set of init parameters (self.flat_popt)
            plotfits (bool):
                if True (default), plot model predictions over observed data
            saveplot (bool):
                if True (default), save plots to model.handler.results_dir
            custompath (str):
                path starting from any subdirectory of "~/" (e.g., home).
                all saved output will write to "~/<custompath>/<self.model_id>/"
            progress (bool):
                track progress across model fits, ninits, and basinhopping
        """
        self.is_nested = True
        self.custompath = custompath
        if flatp is None:
            models = [{'all': 'flat'}] + models
            self.param_sets = self.opt.sample_param_sets()

        pnames = self.toggle_pbars(progress=progress, models=models)
        # loop over depends_on dictionaries and optimize cond models
        for i, depends_on in enumerate(models):
            self.set_fitparams(depends_on=depends_on)
            if progress:
                self.mbar.update(value=i, status=pnames[i])
            if flatp is None:
                self.param_sets = self.opt.sample_param_sets()
                flatp = self.optimize_flat(self.param_sets)
                continue
            self.optimize_conditional(flatp)
            if plotfits:
                self.plot_model_fits(save=saveplot)
            if saveresults:
                self.handler.save_results(saveobserved=saveobserved)
        if progress:
            self.mbar.clear()
            self.opt.gbar.clear()
            self.opt.ibar.clear()


    def recover_model(self, popt=None, yhat=None, nsamples=None, ninits=None, plotparams=True, plotfits=False, progress=False):
        """ fit model to synthetic data similar to observed y-vector
        and compare init params to recovered params to test model identifiabilty
        """
        # set nlevels so simulator result has "conditional" shape
        self.set_fitparams(force_conditional=True)
        if popt is None:
            popt = self.popt
        if yhat is None:
            yhat = self.simulate(p=popt, set_observed=True)
        if nsamples is None:
            nsamples = self.basinparams['nsamples']
        if ninits is None:
            ninits = self.basinparams['ninits']
        nkeep = self.basinparams['ninits']+1
        self.sample_param_sets(nsamples=nsamples, nkeep=nkeep)
        self.optimize(progress=progress, plotfits=plotfits)
        if plotparams:
            # plot init_params agains optimized param estimates
            vis.compare_param_estimates(popt, self.popt, self.depends_on)


    def log_fit_info(self, finfo=None, popt=None, yhat=None):
        """ write meta-information about latest fit
        to logfile (.txt) in working directory
        """
        self.opt.log_fit_info(finfo, popt, yhat)


    def write_results(self, finfo=None, popt=None, yhat=None):
        """ logs fit info to txtfile, fills yhatdf and fitdf
        """

        finfo, popt, yhat = self.set_results(finfo, popt, yhat)
        self.log_fit_info(finfo, popt, yhat)

        self.yhatdf = self.handler.fill_yhatdf(yhat=yhat, fitparams=self.fitparams)
        self.fitdf = self.handler.fill_fitdf(finfo=finfo, fitparams=self.fitparams)
        self.poptdf = self.handler.fill_poptdf(popt=popt, fitparams=self.fitparams)


    def plot_model_fits(self, y=None, yhat=None, kde=True, err=None, save=False, bw='scott', savestr=None, same_axis=True, clrs=None, lbls=None, cumulative=True, simdf=None, suppressLegend=False, simData=None, condData=None, shade=True, plot_error_rts=True, figure=None):
        """ wrapper for radd.tools.vis.plot_model_fits
        """
        data = self.handler.data.copy()

        if y is None:
            y = self.fitparams.y

        if yhat is None:
            try:
                yhat = self.yhat
            except AttributeError:
                print("No Model Predictions to Plot (need yhat argument)")
                yhat = analyze.rangl_data(simdf, quantiles=self.quantiles)
        if save:
            if savestr is None:
                savestr = self.fitparams.model_id
            if self.fitparams['fit_on']=='subjects':
                savestr = savestr + str(self.fitparams['idx'])

        if err is None:
            err = self.handler.observed_err

        if lbls is None and self.fitparams.nlevels>1:
            from itertools import product
            levels = [self.clmap[cond] for cond in self.conds]
            level_data = list(product(*levels))
            lbls = ['_'.join([str(lvl) for lvl in lvls]) for lvls in level_data]

        if self.ssd_method == 'central':
            ssd = self.ssdDF.groupby(self.conds).mean()[0].values
            ssderr = self.ssdDF.groupby(self.conds).sem()[0].values
        else:
            ssd = self.fitparams.ssd_info[0]
            ssderr = None

        fig = vis.plot_model_fits(y, yhat, err=err, quantiles=self.quantiles, ssd=ssd, ssderr=ssderr, bw=bw, same_axis=same_axis, clrs=clrs, lbls=lbls, cumulative=cumulative, save=save, savestr=savestr, suppressLegend=suppressLegend, shade=shade, plot_error_rts=plot_error_rts, figure=figure)


    def simulate(self, p=None, analyze=True, set_observed=False, raw_data=False):
        """ simulate yhat vector using
        :: Arguments ::
            p (dict):
                parameters dictionary
            analyze (bool):
                if True (default) returns yhat vector. else, returns decision traces
        :: Returns ::
            out (array):
                1d array if analyze is True, else ndarray of decision traces
        """
        if p is None:
            try:
                p = self.popt
            except Exception:
                p = self.__get_default_inits()
        p = deepcopy(p)
        yhat = self.sim.sim_fx(p, analyze=analyze)
        if set_observed:
            yhat = yhat.reshape(self.observed[0].shape)
            self.observed = [yhat]
            self.observed_flat = [yhat.mean(axis=0)]
        elif raw_data:
            p = self.sim.vectorize_params(deepcopy(p))
            return self.make_simulated_dataframe(yhat[0], yhat[1], p)
        return yhat


    def make_simulated_dataframe(self, DVg, DVs, p):
        """ get rt and accuracy of go and stop process for simulated
        conditions generated from simulate_dpm
        """
        ssd, nssd, nss, nss_per, ssd_ix = self.sim.ssd_info
        nl, ntot = self.sim.nlevels, self.sim.ntot
        gdec = self.sim.go_resp(DVg, p['a'])
        # if dpm, simply ss_resp() uses 0
        # as boundary simply ignores sec. arg
        sdec = self.sim.ss_resp(DVs, p['a'])
        gort = self.sim.go_RT(p['tr'], gdec)
        ssrt = self.sim.ss_RT(ssd, sdec)
        ert = gort[:, :nss].reshape(nl, nssd, nss_per);
        gort[np.isnan(gort)] = self.fitparams['tb'] + self.sim.dt
        ssrt[np.isnan(ssrt)] = self.fitparams['tb'] + self.sim.dt
        return pandaify_results(gort, ssrt, clmap=self.clmap, ssd=ssd)
